package Puma::Parser;
use ibase Puma::Core;

=head1 NAME

Puma::Parser - the base Puma tag parser

=head1 SYNOPSIS

No synopsis available at this time.

=head1 DESCRIPTION

The Puma tag parser is the core module for all tag parsing functions.
It is used as a base class for Puma::Parser::XML and Puma::ServerPages::Parser.

At this time, the API documentation for this class is still limited, given
the push to get to a functional release.
If you wish to extend this class, check out the two modules listed above and
the rest of this documention.
Good luck!

=cut

use Puma::RegEx qw(getRegEx);  # to retrieve the more complex regular expressions
use Puma::Util qq(readfile);   # to read the file in a single chunk

use strict;

# --- standard constructor ---
sub construct {
	my ($self, %args) = @_;

	# --- get the rules for extracting valid labels ---
	$self->stripArgs(\%args, qw(LabelRules Handlers IncludeTag Text FileName));

	return $self->SUPER::construct(%args);
}

=head1 METHODS

=head2 parse() parseFile(FILENAME)

parse() can be called in any of the following methods:

   parse(TEXT)
   parse(Text => TEXT)
   parse(FileName => FILENAME)

It returns a data structure representation of the parsed text that looks
like this:

   $tree = [ [
      label,
      {
         param => [ value, ... ],
         ...
      },
      text | [ tree ]
   ] ... ];

This data structure is similar to the structure created by the (CPAN)
XML::Parser package.

parseFile() is simply a wrapper around parse() that takes the name of a file
as a parameter and invokes the parser as parse(FileName => FILENAME);

=cut

sub parse {
	my $self = shift;
 
	# --- process incoming parameters ---
	unshift(@_, 'Text') if @_ == 1; my %args = @_;  # presume 'Text' if only 1 arg
	$self->stripArgs(\%args, qw(LabelRules Handlers IncludeTag Text FileName));

	# --- load the text to be parsed ---
	delete $self->{text};
	$self->{text} = $self->{Text} if $self->{Text};
	if ($self->{FileName}) {
		$self->{text} = readfile($self->{FileName});
		print "error reading file '$self->{FileName}'\n" unless $self->{text};
	}
	delete @{$self}{qw(Text FileName)};
	return undef unless $self->{text};

	# --- build the root tag ---
	my $tag = {
		label  => ':parser',
		type  => 'root',
		params => {},
	};

	# --- grab the regular expression for finding the tags ---
	# --- store rules in $self so we only have to grab them once ---
	$self->{_rules} = getRegEx('TagParser', $self->{LabelRules});

	# --- recurse! ---
	$self->{_depth} = 0;
	my $tree = $self->_buildTree($tag);
	delete $self->{_depth};

	return $tree;
}

sub parseFile {
	my ($self, $filename) = @_;
	return $self->parse(FileName => $filename);
}

#	buildTag - a helper function to build a tag structure
#
#	This method takes two or more parameters and returns a tag structure in the
#	format specified above.
#
#	The two parameters are the tag label and a hasshref of tag paramters.  All
#	following fields are text or references to child tags.

sub buildTag {
	my ($self, $label, $params, @children) = @_;
	return [ $label, $params || undef, @children ];
}

# === Private Methods =======================================================

#	_buildTree - Build the tag tree (structure)
#
#	This method calls itself recursively to build a tree of tags.

sub _buildTree {
	my ($self, $tag) = @_;

	# --- don't recurse more than 40 levels deep ---
	return if $self->{_depth} > 40;
	$self->{_depth}++;

	# --- use precompiled rules to search through text and build tag tree ---
	my @children;
	TAG: while ($self->{text} =~ m/$self->{_rules}/gc) { # '/o' to compile regex once
		#! TODO: The /o will not work if the parser is used for both Config and SP
		#!       Figure out a way to compile once per instance

		push @children, $1 if $1;  # push text as a child

		my $child = $self->_splitTag($2);

		# --- check handlers for matching tag label ---
		foreach my $key (keys %{$self->{Handlers}}) {
			if ($child->{label} =~ /$key/) {
				my @handlertags = &{$self->{Handlers}->{$key}}($self, $child);
				push @children, @handlertags if @handlertags;
				next TAG;
			}
		}

		if ($child->{type} =~ /selfending/) {
			push @children, $self->buildTag($child->{label}, $child->{params});
		} elsif ($child->{type} eq 'beginning') {
			push @children, $self->_buildTree($child);
		} elsif ($child->{type} eq 'ending') {
			last if $child->{label} eq $tag->{label} ||
				$self->endingtest($child->{label}, $tag->{label});
		}
		#! TODO: else generate mismatched tag error
	}

	# --- grab text between last tag and EOF ---
	if ($self->{_depth} == 1) {
		$self->{text} =~ m/\G(.*)/gs;
		push @children, $1 if $1;  # push text as a child
	}

	# --- return parsed tree ---
	$self->{_depth}--;
	return $self->buildTag($tag->{label}, $tag->{params}, @children);
}

#	_splitTag - Breaks the tag into a label and key/value pairs

sub _splitTag {
	my ($self, $text) = @_;

	# --- determine the tag type ---
	my $tag = $self->prehandler($text);
	unless (defined $tag) {
		$tag = {
			type  => 'beginning',
			label => '',
			class => 'normal',
		};

		SWITCH: {
			# --- in-line code ---
			($text =~ /^<:\s/) && do {
				$tag = {
					type  => 'selfending',
					label => ':',
					class => 'code',
				};
				last;
			};
			# --- in-line code returning a display value ---
			($text =~ /^<:=\s/) && do {
				$tag = {
					type  => 'selfending',
					label => ':=',
					class => 'eval',
				};
				last;
			};
			# --- self-ending tag ---
			($text =~ m|/>$|) && do { $tag->{type} = 'selfending'; last; };
			# --- ending tag ---
			($text =~ m|^</|) && do { $tag->{type} = 'ending'; last; };
		}
	}

	# --- process code and eval tags ---
	if ($tag->{class} =~ /code|eval/) {
		#! TODO: detect and throw error when tag not closed properly.
		$self->_addTagParam($tag, $tag->{class}, ($text =~ m|<$tag->{label}\s(.*?)/>$|s)[0]);
		return $tag;
	}
	# --- process code and eval tags ---
	if ($tag->{class} =~ /raw/) {
		#! TODO: detect and throw error when tag not closed properly.
		$self->_addTagParam($tag, text => ($text =~ m|<$tag->{label}\s(.*?)>$|s)[0]);
		return $tag;
	}

	# === process all other (not code or eval) tags ===
	# --- get the label ---
	$text =~ m|</?([^>\s/]+)\s*(?:/?>)?|g;
	$tag->{label} = $1 if $1;

	# --- load the regexs for grepping the tag ---
	my $pairregex = getRegEx('TagPair');
	my $loneregex = getRegEx('TagStandAlone');

	my $lastparam = 0;
	while (!$lastparam) {
		if ($text =~ m/$pairregex/gco) {

			# --- found a paired value ---
			my ($key, $value) = ($1, $+);
			$value =~ s/\\([`'"])/$1/gs;  # normalize escapes
			$self->_addTagParam($tag, $key, $value);

		} elsif ($text =~ m/$loneregex/gco) {

			# --- found a key with no value ---
			$self->_addTagParam($tag, $+, '');

		} else {

			# --- we've hit the last parameter ---
			$lastparam = 1;
		}
	}

	return $tag;
}

#	_addTagParam - helps build the tag parameter list

sub _addTagParam {
	my ($self, $tag, $key, $value) = @_;

	my $params = $tag->{params};
	if (exists $params->{$key} && defined $params->{$key}) {
		if (ref $params->{$key} eq 'ARRAY') {
			push @{$params->{$key}}, $value;
		} else {
			$params->{$key} = [ $params->{$key}, $value ];
		}
	} else {
	   $params->{$key} = $value;
	}

	$tag->{params} = $params;
}

# --- tag prehandler (see splitTag) ---
# --- prehandler stub ---
# incoming: tag text (< to >)
# returns: a teg hash or undef if not processed
# set $tag->{type} = 'raw' to stop name-value parsing
sub prehandler { return undef; }
sub endingtest { return undef; }

#	Intermediate tag structure
#
#	$tag = {
#		type  => beginning | ending | selfending | root
#		label => prefix:name
#		class => normal | code | eval | raw
#	};


=head1 SEE ALSO

Puma::RegEx(3)
Puma::Parser::XML(3)
Puma::ServerPages::Parser(3)
XML::Parser

=head1 COPYRIGHT

Copyright (c) 2001, 2002 by Ingmar Ellenberger.

Distributed under The Artistic License.  For the text of this license,
see http://puma.site42.com/license.psp or read the file LICENSE in the
root of the distribution.

=cut

1;
