package Puma::RegEx;
use ibase Exporter;

=head1 NAME

Puma::RegEx - regular expression repository for Puma

=head1 SYNOPSIS

 use Puma::RegEx qw( getRegEx );
 my $regex = getRegEx('regexname', options ... );

=head1 DESCRIPTION

The Puma::RegEx package returns pre-defined regular expressions for Puma's
core functions.

By placing many of the more complex regular expressions in this package we
have managed to make the original code more readable through the elimination
of heredocs and allowing room for more extensive documentation of the regular
expressions.

=head1 FUNCTIONS

=head2 getRegEx(NAME, [OPTIONS], ...]])

This method returns a compiled regular expression based on the parameters
passed to it.
The OPTIONS passed to the function vary based on the NAME of the regular
expression requested:

'tagparser'

The 'tagparser' regex is the main regex for separating markup out of the
data stream.
It takes an aditional parameter 'LABELRULES' that defines additional rules
so that the regex can identify non-standard or proprietary tags.

The regex returns two values ($1 and $2) when run: the text before the tag
and the content of the tag itself.

'tagpair' and 'tagstandalone'

These regular expressions extract values from within a tag.

'tagpair' breaks out a name="value" pair into $1 and $2.
'tagstandalone' returns the content of a standalone tag as $1.

=cut

@EXPORT_OK = qw(getRegEx);


sub getRegEx {
	my $name = lc(shift);

	my $regex = undef;
	SWITCH: {

		# --- regex for detecting and extracting tags ---
		'tagparser' eq $name && do {
			my $labelrules = shift;
			$labelrules = $labelrules ? '(?:'. $labelrules .')' : '';

			$regex =
<<'REGEX';

	# --- grab text before the tag ($1)---
	(.*?)

	# --- start matching tag ($2)---
	(<  # opening brace
		(?:
			(:[\w=]*\s)         # match a code or eval tag ($3)
			| LABELRULES  # or the label rules (substituted in later)
		)

		# --- test for quotes so we don't wrongly detect tag endings ---
		(?:
			# --- test for doublequoted string ---
			"                 # opening quote
			# TODO: can the next 6 lines be shortened to (?:[^"\\]|\\.)* ???
			(?:
				[^"\\]*        # grab everything that's not a quote or backslash
				(?:
					\\.[^"\\]*  # found a backslash
				)*
			)
			"                 # closing quote

			# --- or a singlequoted string ---
			|'(?:[^'\\]*(?:\\.[^'\\]*)*)'

			# --- or a backquoted string ---
			|`(?:[^`\\]*(?:\\.[^`\\]*)*)`

			# --- or anything else ---
			|.*?
		)
		# --- ... and do it repeatedly ---
		*?

		# --- if code tag found ($3 matched), require self-ending tag ---
		(?(3)/)

	# --- end of tag match ---
	>)  # closing brace

REGEX
			$regex =~ s/LABELRULES/$labelrules/s;
			$regex = qr/\G$regex/sx;  # '/x' filters comments and whitespace

			last;
		};

		# --- regex for splitting tags pairs ---
		'tagpair' eq $name && do {
			# TODO: add comments to regex
			$regex =
<<'REGEX';

	([\w.-]+)\s*=\s*              # get key and '='
	(?:"([^"\\]*(?:\\.[^"\\]*)*)" # doublequoted value
	  |'([^'\\]*(?:\\.[^'\\]*)*)' # singlequoted value
	  |(`[^`\\]*(?:\\.[^`\\]*)*`) # backquoted value (keep quotes)
	  |([^\s>]*)                  # unquoted value
	)\s*                          #    and the space afterward

REGEX
			$regex = qr/\G$regex/sx;  # '/x' filters comments and whitespace
			last;
		};

		# --- regex for getting a standalonetags ---
		'tagstandalone' eq $name && do {
			# TODO: add comments to regex
			$regex =
<<'REGEX';
	(?:([\w.-]+) # get key
	  |(`[^`]*`) #    or an eval
	)\s*
REGEX
			$regex = qr/\G$regex/sx;  # '/x' filters comments and whitespace
			last;
		};
	};

	return $regex;
}

=head1 COPYRIGHT

Copyright (c) 2001, 2002 by Ingmar Ellenberger.

Distributed under The Artistic License.  For the text of this license,
see http://puma.site42.com/license.psp or read the file LICENSE in the
root of the distribution.

=cut

1;
