package Puma::ServerPages::Config;
use ibase Exporter;

=description

	Function to build a configuration hash for a ServerPages file

=cut

@EXPORT_OK = qw( getConfig );

use Puma::Config;
use Puma::Util 'mergehash';
use Puma::Util::XML 'xml2hash';

use strict;

=Configuration File Rules of Conduct

	<etc>     = /etc/puma
	<webroot> = DOCUMENT_ROOT
	<uri>     = REQUEST_URI
	<cwd>     = PATH_TRANSLATED minus the filename

	cascade="off"
		Only processes <etc>/config.xml

	cascade="root"
		Only processes <webroot>/.config.xml and <etc>/config.xml

	cascade="local"
		Processes <cwd>/.config.xml and all 'parent' references, then same as
		"root".  Will not search path for another .config.xml if there is none in
		<cwd>.

	cascade="on" (default)
		Same as "local" except that it looks up the fs tree for a .config.xml
		when there is not one in <cwd>.  It will not search the tree unless
		<webroot>/<uri> is not the same as <cwd>.

	cascade="full"
		Same as "on" except that it will look up the filesystem tree even if
		<webroot>/<uri> is not the same as <cwd>.

	file=".config.xml" <default>
		Used in the <etc>/config.xml, this sets/changes the name of the
		configuration file that this module looks for.

	parent="dir" or "file"
		Essential for cascading, this directive in the .config.xml files
		indicates which parent directory to look in for the next cascading XML
		file.

=cut

sub getConfig {
	# --- start with the /etc/puma/config file ---
	my $etc = Puma::Config->getConfig || {};
	$etc = $etc->{puma}->{serverpages} || {};

	# --- load the default settings ---
	my $cascade = $etc->{config}->{cascade} || 'on';
	my $file    = $etc->{config}->{file}    || '.config.xml';
	my $parent  = $etc->{config}->{parent}  || '';

	# --- cascade="off" ---
	return $etc if $cascade eq 'off';

	# --- load the config file form the document root 
	#     and create the @configs array of hashrefs ---
	#! TODO: use _loadconfigfile
	my $root = xml2hash(FileName => "$ENV{DOCUMENT_ROOT}/$file") || {};
	$root = $root->{serverpages} || {};

	# --- cascade="root" ---
	return mergehash($etc, $root) if $cascade eq 'root';

	# --- resolve a few things ---
	my $webroot = $ENV{DOCUMENT_ROOT};
	my $uri = $ENV{REQUEST_URI};
	my $cwd = ($ENV{PATH_TRANSLATED} =~ m|^(.*)/[^/]+$|)[0];

	# === Generate an array of config hashes of .config.xml files ===
	my (@configs, @files);

	# --- look for the first file ---
	if (-e "$cwd/$file") {
		# --- found a file in $cwd! ---
		push @files, "$cwd/$file";
	} elsif ($cwd eq "$webroot/$uri" || $cascade eq 'full') {
		# --- search up tree ---
		my @path = split /\//, $cwd;
		pop @path; # take the last directory off the path
		while (my $path = join '/', @path) {
			if (-e "$path/$file") {
				push @files, "$path/$file";
				last;
			}
			pop @path; # take the last directory off the path
		}
	}
	# --- no additional .config.xml files found ---
	return mergehash($etc, $root) unless @files;

	# === Read the file, stack the config and follow the parents ===
	while (1) {
		# --- load the config file and push it on the stack ---
		my $config = _loadconfigfile($files[0]);
		unshift @configs, $config;

		# --- get the parent filename ---
		my $parent = $config->{config}->{parent}
			if exists $config->{config} && exists $config->{config}->{parent};

		# --- pop out of loop if wa can't find the parent ---
		last unless $parent;
		$parent .= "/$file" unless -f $parent; # tack on $file?
		last unless -f $parent;

		# --- found the file, add to stack ---
		unshift @files, $parent;
	}

	# --- merge and return ---
	return mergehash($etc, $root, @configs);
}

# === Local functions =======================================================

# --- load and do a bit of preprocessing on .config.xml files ---
use Puma::Util::FileSystem 'cleanPath';
sub _loadconfigfile {
	my $file = shift;

	# --- file must exist and be fully qualified ---
	return {} unless -f $file && $file =~ /^\//;

	# --- resolve a few things ---
	my $cwd = ($file =~ m|^(.*)/[^/]+$|)[0];

	# --- load and parse the file ---
	my $hash = xml2hash(File => $file) || {};
	# --- extract the right sub-hash ---
	if (exists $hash->{puma}) {
		$hash = $hash->{puma}->{serverpages} || {}; # for /etc/puma/config.xml
	} else {
		$hash = $hash->{serverpages} || {}; # for all .config.xml files
	}

	# --- expand the <include root="..."/> entry ---
	$ENV{INCLUDE_ROOT} ||= $ENV{DOCUMENT_ROOT};
	$ENV{INCLUDE_URI}  ||= $ENV{REQUEST_URI};
	if (exists $hash->{include}) {
		if (exists $hash->{include}->{root}) {
			my $root = $hash->{include}->{root};
			# --- fill relative paths ---
			$root = "$cwd/$root" unless $root =~ /^\//;
			# --- clean it up and ship it back ---
			$ENV{INCLUDE_ROOT} = $hash->{include}->{root} = cleanPath($root);
		}

		# --- set the include URI root in the ENV hash ---
		$ENV{INCLUDE_URI} = $hash->{include}->{uri}
			if exists $hash->{include}->{uri};
	}

	# --- expand the <config parent="..."/> entry ---
	if (exists $hash->{config} && exists $hash->{config}->{parent}) {
		my $parent = $hash->{config}->{parent};
		# --- fill relative paths ---
		$parent = "$cwd/$parent" unless $parent =~ /^\//;
		# --- clean it up and ship it back ---
		$hash->{config}->{parent} = cleanPath($parent);
	}

	# --- we is be done! ---
	return $hash;
}

1;

=config.xml file

<serverpages>
	<include root="." (other global default include shit) />

	<!-- adds to perl library path -->
	<perl libs="/web/cvs/build/perl" />

	<!-- Creates global vars as $psp->{key} = "value" -->
	<psp
		RootURI="/new"
		TmpDir="/web/tmp"
		BuildDest="/built"
		VSDir="/web/cvs"
		CVSRoot="/cvs"
	/>

	<!-- Sets values in %ENV -->
	<env
		HOME="/home/build"
		PATH="/usr/local/bin/jdk1.3.1/bin:/usr/local/bin:/bin:/usr/bin"
	/>
</serverpages>

=cut


