package Puma::ServerPages::Parser;
use ibase Puma::Parser;

=copyright

	Puma - Perl Universal Markup
	Copyright (c) 2001, Ingmar Ellenberger.  All rights reserved.
	Software is distributed AS IS.  No warranties, expressed or implied.

	This license is temporary, pending the selection of an appropriate
	open-source license.

=description

	ServerPages.pm - Server Pages extension of the Puma Parser

=cut

use Puma::Util qw( quote );

use strict;

# === Object Construction ===================================================
sub construct {
	my ($self, %args) = @_;

	# --- parameter processing ---
	# 'Config' = hash of config data
	$self->stripArgs(\%args, qw( Config Server ));
	$self->{Config} = $self->{Server}->getConfig() if $self->{Server};

#! TODO: move this functionality to Util::quote
#	# --- qqmask always contains a ']' ---
#	$self->{qqmask} = '\]';

	# --- set the rules for extracting valid labels ---
	return $self->SUPER::construct(%args,
		LabelRules => '[^<>\s]*:',
		Handlers => {
			':include' => \&include,
		},
	);
}

# === Tree Codification =====================================================
# --- turn the hash into eval'able code ---
sub codify {
	my ($self, $tree) = @_;

	# --- create the code header ---
	#! TODO: remove extra <lf>s
	my $code = qq[use Puma::ServerPages::Global; use strict;\n\n];

	# --- codify the tree ---
	$code .= $self->_walk(@$tree) if $tree;

	# --- return the generated code ---
	return $code;
}

# --- private method called to recursively walk the tree ---
sub _walk {
	my ($self, $tagname, $params, @tree) = @_;

	# --- split tagname at the colon ($prefix:$name) ---
	$tagname =~ /([^:]*):(.*)/;
	my ($prefix, $name) = ($1 || '', $2 || '');

	# --- start building the code ---
	my $code = '';
	if ($prefix) {

		# --- if we have a prefix, it's a method call ---
		$code .= '$'. $prefix ."->_render( sub { \$$prefix->$name(";

		my $paramstr = '';
		my $needclean = 0;
		foreach my $key (keys %$params) {
			my $value = $params->{$key};

			# --- quote the value unless it's an eval (backquotes) ---
			if ($value =~ /^`([^`]+)`$/) {
				$value = $1;
			} else {
				$value =~ s/(\')/\\$1/g;
				$value = "'$value'";
			}

			if ($key =~ /^`([^`]+)`$/) {
				$key = "($1)"; #|| ' '";
				$needclean = 1;
			}

			$paramstr .= " $key => $value,";
		}
		if ($needclean) {
			$code .= "cleanhash($paramstr)";
		} else {
			$code .= $paramstr;
		}
		$code .= ") }, sub {\n";
	} else {

		# --- if there is no prefix, it's a special call or a dud ---
		return "$params->{code}\n" unless $name; # in-line code
		return "$params->{code}\n" unless $name; # in-line code
		#unless ($name) {
		#	return "$params->{code}\n" if $params->{code};
		#	return;
		#}

		return qq[print($params->{eval});\n] if $name eq '='; # evaluations

		# --- process special commands - '<:command ...' ---
		SWITCH: {

			# --- load an object as an application ---
			$name =~ /^(?:app|use)$/ && do {
				$code .= "use $params->{module}; my \$$params->{prefix} = new $params->{module}(";
				foreach my $key (keys %$params) {
					my $value = $params->{$key};

					# --- quote the value unless it's an eval (backquotes) ---
					if ($value =~ /^`([^`]+)`$/) {
						$value = $1;
					} else {
						$value =~ s/(\')/\\$1/g;
						$value = "'$value'";
					}

					$code .= "'$key'=>$value," unless $key =~ /module|prefix/;
				}
				$code .= "Server=>\$server);\n";
				last;
			};

			# --- add library to path ---
			$name eq 'lib' && do {
				$code .= "use lib '$params->{path}';\n\n";
				last;
			};

			# --- process interpolation tag ---
			$name eq 'interpolate' && do {
				if (@tree) { $params->{type} = 'beginning'; }
					else     { $params->{type} = 'selfending'; }

				$self->_interpolate(%$params);
				last;
			};

			# --- processing for conditionals and loops ---
			$name =~ /^elsif$/ && do {
				my $value = $params->{text};
				$value =~ s|\s*/$||g;

				$code .= "} $name $value { ";
				last;
			};
			$name =~ /^(?:if|for|foreach|unless|until|while)$/ && do {
				my $value = $params->{text};
				$value =~ s|\s*/$||g;

				$code .= "$name $value { ";
				last;
			};
			$name =~ /^(?:else|continue)$/ && do {
				$code .= "} $name { ";
				last;
			};

			# --- if we get here and it's not the root :parser tag, barf ---
			# (we ignore the :parser tag)
			$name ne 'parser' && do {
				$code .= qq[print("<b>Unknown Tag '$tagname'</b><br>\\n");\n];
				last;
			};
		}

	}

	if (@tree) {
		# --- walk the tag array ---
   	foreach my $child (@tree) {
			if (ref $child) {
				$code .= $self->_walk(@$child);
			} else {
				$code .= 'print '. quote($child) .";\n";
			}
		}

		# --- process closing interpolation tag ---
		if (!$prefix && $name eq 'interpolate') {
			$params->{type} = 'ending';
			$self->_interpolate(%$params);
		}
	}

	if ($prefix) {
		$code .= "});\n";
	} elsif ($name) {
		SWITCH: {
			$tagname =~ /^:(?:if|for|foreach|unless|until|while)$/ && do {
				$code .= '} ';
				last;
			};
		}
	}

	return $code;
}

# === Parent Method Overrides ===============================================
# --- prehandler (used for conditional and loop tags) ---
sub prehandler {
	my ($self, $text) = @_;

	my $tag;

	if ($text =~ /^<:(if|for|foreach|unless|until|while)[\s\(>]/) {
		$tag = {
			type  => 'beginning',
			label => ":$1",
			class => 'raw',
		};
	}
	if ($text =~ /^<:(continue|else|elsif)[\s>]/) {
		$tag = {
			type  => 'selfending',
			label => ":$1",
			class => 'raw',
		};
	}	
	if ($text =~ /^<\/:\s*/) {
		$tag = {
			type  => 'ending',
			label => '/:',
			class => 'raw',
		};
	}

	return $tag;
}

# --- match and resolve closing tags ---
sub endingtest {
	my ($self, $endtag, $starttag) = @_;

	return 1 if $endtag eq '/:' && $starttag =~ /:(?:if|for|foreach|unless|until|while)/;
	return undef;

}

# === Private Methods =======================================================

#! TODO: This likely doesn't work any more - needs a re-write

=method
	_interpolate - Turn on/off in-line variable interpolation

	<:interpolate on|off mask='&%@$'/>
=cut

sub _interpolate {
	my ($self, %params) = @_;

	my $mask;
	SWITCH: {
		defined $params{mask} && do { $mask = $params{mask}; last; };
		defined $params{on}   && do { $mask = '';            last; };
		defined $params{off}  && do { $mask = '&%@$';        last; };
	}
	# --- mask always contains a ']' ---
	$mask .= '\]';

	SWITCH: {
		# --- presume self ending if no type ---
		!defined $params{type} && last;

		$params{type} eq 'beginning' && do {
			my $istack = $self->{qqstack} || [];
			push @$istack, $self->{qqmask};
			$self->{qqstack} = $istack;
			last;
		};
		$params{type} eq 'ending' && do {
			last unless ref $self->{qqstack};
			$mask = pop @{$self->{qqstack}};
			last;
		};
	}

	$self->{qqmask} = $mask;
	return;
}

#! TODO: Whatthehell are these two methods for?  I'm sure I put them here for a reason.
# --- evaluate the resulting code ---
sub evaluate { my $self = shift; return eval(shift); }
sub write { my $self = shift; }


# === Server-side Include ===================================================

=method

	include - Compile time server-side include

	Added as a part of the new Handler mechanism for the parser, this method
   allows server-side includes whenever the ':include' label is found.

=cut

use Puma::ServerPages::Include;
sub include {
	my ($self, $tag) = @_;
	return Puma::ServerPages::Include->include(
		Config => $self->{Config},
		%{$tag->{params}}
	);
}

1;
