package Puma::ServerPages::Server;

=copyright

	Puma - Perl Universal Markup
	Copyright (c) 2001, Ingmar Ellenberger.  All rights reserved.
	Software is distributed AS IS.  No warranties, expressed or implied.

	This license is temporary, pending the selection of an appropriate
	open-source license.

=description

	Server.pm - Server object for Puma::ServerPages

=cut

# Combine various elements from CGI, %ENV and other sources to assemple a
#   server boject similar to the ASP $Server object.
use ibase 'Puma::Core';

use CGI;
use CGI::Carp 'fatalsToBrowser';
use Puma::Util 'quote';
use Puma::Util::XML 'xml2hash';

# === Constructor ===========================================================

sub construct {
	my $self = shift;

	# --- call superconstructor and parse local params ---
	my %args = $self->SUPER::construct(@_);
	$self->stripArgs(\%args, qw[
		Config XML
	]);

	#! TODO: read config file here
	if (defined $self->{Config} && ref $self->{Config} eq 'HASH') {
		# we have a valid configuration hash, so do nothing
	} elsif (defined $self->{XML}) {
		$self->{Config} = xml2hash(FileName => $self->{XML})->{puma};
	}

	# --- create the CGI object ---
	$self->{CGI} = new CGI;

	# --- do the cookie thing ---
	my $cookies = $self->{Config}->{cookie};
	$cookies = [ $cookies ] unless ref($cookies) eq 'ARRAY';

	for (my $ii = 0; $ii < @$cookies; $ii++) {
		$self->{message} .= " bake ". Data::Dumper->Dump([$cookies->[$ii]], ['cookie']);
		$self->bake($cookies->[$ii]);
	}

	return %args;
}

# === Generate code for top and bottom of page ==============================

sub getStartCode {
	my $self = shift;
	my $config = $self->getConfig;

	my $sessions = $config->{session};
	$sessions = [ $sessions ] unless ref($sessions) eq 'ARRAY';

	my $code = '';
	# --- create the code to load the session objects ---
	foreach my $session (@$sessions) {
		# --- load the session object module ---
		$code .= qq[ use $session->{module}; ];

		# --- instantiate the session object ---
		#! TODO: use the $include->{root} for the location unless it's an absolute path
		$code .= qq[
			my \$$session->{prefix} = new $session->{module}(
				Location  => '$session->{Location}',
				Context   => '$session->{Context}',
				UseCookie => '$session->{UseCookie}',
				Server    => \$server,
			);
		] if $session->{prefix};
	}

	# --- get rid of extra whitespace ---
	$code =~ s/^\s+//g;
	$code =~ s/\s+/ /g;

	# --- load libs ---
	$code .= "\nuse lib qw($config->{perl}->{libs});" if $config->{perl}->{libs};

	# --- set the global $psp hash ---
	$code .= "\nmy \$psp = \$server->getConfig()->{psp};";
	
	# --- set local vars ---
	if ($config->{vars}) {
		$code .= "\n";
		foreach my $key (keys %{$config->{vars}}) {
			$code .= "my \$$key = ". quote($config->{vars}->{$key}) .'; ';
		}
	}

	# --- run startcode ---
	$code .= "\n$config->{startscript}\n" if $config->{startscript};

	return $code;
}

sub getEndCode {
	my $self = shift;
	my $config = $self->getConfig;

	my $sessions = $config->{session};
	$sessions = [ $sessions ] unless ref($sessions) eq 'ARRAY';

	my $code = '';
	# --- run endcode ---
	$code .= "\n$config->{endscript}\n" if $config->{endscript};

	# --- close off sessions ---
	foreach my $session (@$sessions) {
		$code .= qq[\$$session->{prefix}->save();\n];
	}

	return $code;
}

# === Cookie Handling =======================================================

sub bake {
	my $self = shift;

	my %args = @_ == 1 ? %{$_[0]} : @_;

	my $name    = $args{name};
	my $expires = $args{expires};

	use CGI::Cookie;
	my %cookies = fetch CGI::Cookie;

	my $cookie;
	if (exists $cookies{$name} && defined $cookies{$name}) {
		$cookie = $cookies{$name};
		$cookie->expires($expires) if $expires;
	} else {
		my $value = crypt(time * rand, 'sp');
		$value =~ s/[^0-9A-Za-z]//g;
		my %params = (
			-name  => $name,
	   	-value => $args{value} || $value, # crypt(time * rand, 'sp'),
		);
		$params{'-expires'} = $expires if $expires;

		$cookie = new CGI::Cookie(%params);
	}

	$self->{_baked} = [] unless $self->{_baked};
	push @{$self->{_baked}}, $cookie;

	return $cookie;
}

sub getCookie {
	my ($self, $name) = @_;

	return undef unless $name;

	foreach my $cookie (@{$self->{_baked}}) {
		return $cookie if $cookie->name() eq $name;
	}

	return undef;
}

# === CGI Pass-thrus ========================================================
sub param { my ($self, @args) = @_; return $self->{CGI}->param(@args); }

# === CGI Overrides =========================================================
sub header {
	my ($self, @args) = @_;

	return $self->{CGI}->redirect($self->redirect) if $self->redirect;

	push @args, '-cookie' => $self->{_baked} if $self->{_baked};
	return $self->{CGI}->header(@args);
}

#! TODO: Make a hashtie class that does an auto array cast on values & holds order

# === CGI Augmentation ======================================================
sub paramHash {
	my $self = shift;

	my $cgi = $self->{CGI};
	my @keys = $cgi->param;
	my $hash = {};

	foreach my $key (@keys) {
		if (@{$cgi->{$key}} == 1) {
			$hash->{$key} = $cgi->{$key}->[0];
		} else {
			$hash->{$key} = $cgi->{$key};
		}
	}

	return wantarray ? %$hash : $hash;
}

sub redirect {
	my ($self, $url) = @_;

	return $self->{_redirect} unless defined $url;
	return $self->{_redirect} = $url;
}

# --- current working directory ---
sub cwd { return ($ENV{PATH_TRANSLATED} =~ m|^(.*/)[^/]+$|)[0]; }

# === Accessors =============================================================

sub getConfig { return shift->{Config}; }

# === keeeerap! =============================================================
sub test { print "this is a test\n"; }

1;
