package Puma::State::DDState;
use ibase Puma::Core;

=copyright

	Puma - Perl Universal Markup
	Copyright (c) 2001, Ingmar Ellenberger.  All rights reserved.
	Software is distributed AS IS.  No warranties, expressed or implied.

	This license is temporary, pending the selection of an appropriate
	open-source license.

=description

	DDState.pm - A simple object implemented using Data::Dumper

=cut

use Puma::Util qw(readfile writefile);
use Data::Dumper;

use strict;

# TODO: build config object to be passed in
#    $self->{_location} - db, path, whatever.
#    $self->{_context}  - application name
#    $self->{_uid}      - unique session key

# --- object contructor ---
# an override of the Puma::Core constructor is required to properly tie the hash
sub new {
	# --- a hellofan effort to get a blessed obj ref to a tied hash! ---
	my $this = shift;                    # get classname or ref to instantiated object
	my $class = ref($this) || $this;     # get classname from ref if necessary
	my %hashtie = ();                    # create a hash to tie
	tie %hashtie, $class;                # tie hash to class (for hash tie)
	my $self = bless \%hashtie, $class;  # bless ref to tied hash into class (for object)

	# --- constructor call extracted from Puma::Core ---
	eval { $self->construct(@_); };
	warn $@ if $@;

	return $self;
}

sub construct {
	my $self = shift;
	my %args = $self->SUPER::construct(@_);

	# --- Pull in remaining params ---
	$self->reserve(
		_location => $args{Location} || '.',
		_context  => $args{Context}  || 'ddstate',
		_uid      => $args{UID}      || '0',
		_content  => {},
	);

	return %args;
}

# --- load and save data, and generate the filename ---
sub load {
	my $self = shift;
	my $eval = readfile($self->filename);
	eval $eval if $eval;
}

sub save {
	my $self = shift;

	my $dump = new Data::Dumper([$self->{_content}], ['self->{_content}']);
	$dump->Indent(0);
	$dump->Purity(1);

	writefile($self->filename, $dump->Dump);
}

sub filename {
	my $self = shift;

	my $filename = $self->{_location} .'/'. $self->{_context} .'.'. $self->{_uid};
	$filename =~ s|/+|/|g;
	return $filename;
}

# TODO: return epoch seconds since last access
sub lasttouch { return 0; }

# --- generate a regex of reserved words ---
# TODO: combine reservedregex and reservedwords
sub reservedregex {
	my $self = shift;

	my $reserved = $self->reservedwords();
	return qr/$reserved/s;
}

sub reservedwords {
	my $self = shift;

	# --- return words delimited with pipes or '_reserved' ---
	return
		defined $self->{_reserved}
			? join('|', @{$self->{_reserved}})
			: '_reserved';
}

sub reserve {
	my $self = shift;

	# --- make sure we have an even number of arguments ---
	my %args = @_ % 2 ? (@_, '') : @_;

	$self->{_reserved} = [ '_reserved' ] unless defined $self->{_reserved};
	foreach my $key (keys %args) {
		push @{$self->{_reserved}}, $key;
		$self->{$key} = $args{$key};
	}

	# TODO: sort/unique the _reserved array
}

# === hash tie stuff ========================================================

# --- hash constructor ---
sub TIEHASH {
	my ($this, $hash) = (shift, shift || {});
	return bless $hash, ref($this) || $this;
}

sub STORE {
	my ($self, $key, $value) = @_;
	my $reserved = $self->reservedregex();
	return $self->{$key} = $value if $key =~ /$reserved/;
	return $self->{_content}->{$key} = $value;
}

sub FETCH    {
	my ($self, $key) = @_;
	my $reserved = $self->reservedregex();
	return $self->{$key} if $key =~ /$reserved/;
	return $self->{_content}->{$key};
}

sub FIRSTKEY { my $a = scalar keys %{$_[0]->{_content}}; each %{$_[0]->{_content}} }
sub NEXTKEY  { each %{$_[0]->{_content}} }
sub EXISTS   { exists $_[0]->{_content}->{$_[1]} }
sub DELETE   { delete $_[0]->{_content}->{$_[1]} }
sub CLEAR    { %{$_[0]->{_content}} = () }1;

1;
