package Puma::Util;
use ibase Exporter;

=head1 NAME

Puma::Util - various stand-alone utilities

=head1 SYNOPSIS

use Puma::Util qw( readfile writefile
                   splithash mergehash quote );

=head1 DESCRIPTION

Puma::Util provides a number of stand-alone utility functions.
Most of these functions do not have any direct dependancy on Puma per se, but
add greatly to the value of the package my making a number of common
functions quite easy.

This package encalsulates a number of miscellaneous functions that do not fit
into a category in the Puma::Util:: subpackages.

=cut

@EXPORT_OK = qw(
	readfile writefile readFile writeFile
	splithash mergehash
	quote

	lckey ncget ncset fixpath cleanhash

	writelog
);

use Symbol;

use strict;

# === Read a file, return the contents ======================================

=head1 FUNCTIONS

=head2 readfile(FILENAME) or readFile(FILENAME)

The readfile functions opens the file named in the scalar FILENAME
and returns the contents of the file.

readFile() is a synonym for readfile() for those sensitive to case.

N.B.  These functions may be moved to Puma::Util::File (or some similar
class) sime time in the future.

=cut

sub readFile { return readfile(@_); } # for the case sensitive
sub readfile {
	my $filename = shift;

	return undef unless $filename;

	# --- undefine system record separator to read file in one big chunk ---
	my $recordseparator = $/; undef $/;

	# --- open file and read contents ---
	my $fh = gensym();
	open $fh, $filename or return undef;
	my $content = <$fh>;
	close $fh;
	ungensym($fh);

	# --- set system record separator back to original value ---
	$/ = $recordseparator;

	# --- return file contents ---
	return $content;
}

# === Write (create/overwrite) a file =======================================

=head2 writefile(FILENAME, CONTENT) or writeFile(FILENAME, CONTENT)

writefile writes CONTENT to the file named FILENAME without any notice or
warning, overwriting any existing content, truncating the file to exactly the
length of CONTENT or creating a new file if none exists.

writeFile() is a synonym for writefile() for those sensitive to case.

N.B.  These functions may be moved to Puma::Util::File (or some similar
class) some time in the future.

=cut

sub writeFile { return writefile(@_); } # for the case sensitive
sub writefile {
	my $filename = shift;
	my $content = shift || '';

	return unless $filename;

	# --- make the parent directories ---
	my @dirs = split /\//, $filename;
	if ($filename =~ /^\//) {
		shift @dirs;
		$dirs[0] = '/'. $dirs[0];
	}
	my $path;
	for (my $ii = 0; $ii < @dirs - 1; $ii++) {
		$path = ($path ? $path .'/' : ''). $dirs[$ii];
		mkdir $path, 0777;
	}

	# --- (over)write the file ---
	my $fh = gensym();
	open $fh, ">$filename" or return;
	print $fh $content;
	close $fh;
	ungensym($fh);
}

# === Methods for doing funky things with hashes ============================

=head2 splithash(HASHREF, KEYARRAY)

splithash takes HASHREF, removes the key/value pairs listed in KEYARRAY and
returns a new HASHREF with the extracted keys.

It will not create any keys that did not previously exist or were undefined.

=head2 mergehash(HASHREF1, HASHREF2, ... HASHREFn);

This function returns a new hash consisting of the merged contents of all
given HASHREFs.  The HASHREFs that are later in the list will override any
matching values in earlier HASHREFs.

merghash returns an entirely new hash and does not modify the original hashes
or any hashes within them.

N.B.  These functions may be moved to Puma::Util::Hash (or some similar
class) some time in the future.

=cut

# --- split a hash in two by key ---
sub splithash {
	my ($hash1, @keylist) = @_;

	my $hash2 = {};
	foreach my $key (@keylist) {
		if (exists $hash1->{$key} && defined $hash1->{$key}) {
			$hash2->{$key} = $hash1->{$key};
			delete $hash1->{$key};
		}
	}
	return wantarray ? ($hash1, $hash2) : $hash2;
}

# --- overlay one hash (old) on another (new) ---
sub mergehash {
	# --- only given one hash? - return it! ---
	return $_[0] if @_ == 1;

	my $new = {};
	# --- loop through all the hashes merging them one after another ---
	while (@_ > 0) {
		# --- grab hash off stack ---
		my $old = shift;
		# --- push each key on the new hash ---
		foreach my $key (keys %{$old}) {
			if (exists $new->{$key} && ref($new->{$key}) eq 'HASH') {
				# --- recurse if it's another hash ---
				$new->{$key} = mergehash($new->{$key}, $old->{$key});
			} else {
				$new->{$key} = $old->{$key};
			}
		}
	}

	return $new;
}

# === Quote strings without using escapes ===================================

=head2 quote(STRING [, STRING, STRING, ...])

This functions attempts to find a unique quotation delimeter (qq) for
STRING.  It will search through STRING for various forms of delimeters
and attempt to use a delimeter not contained in the string.  If a suitable
delimeter could not be found, it will use an arbitrary quotation delimeter
and escape out any instances of conflicting delimeters.

If quote receives more than one parameter, it will concatentates all
parameters into a single string before processing.

=cut

sub quote {
	my $string = join '', @_;

	my %quotes = (
		'[]' => '\[|\]',
		'()' => '\(|\)',
	);

	foreach my $key (keys %quotes) {
		unless ($string =~ /$quotes{$key}/) {
			return 'qq'. substr($key, 0, 1) . $string . substr($key, 1, 1);
		}
	}
	$string =~ s/([()])/\\$1/g;
	return "qq($string)";
}


# === Under Development =====================================================
# --- a simple utility to write messages to a log file ---
sub writelog {
	my $filename = '/tmp/error_log';

	my $fh = gensym();
	open $fh, ">>$filename";
	chmod 0666, $filename;
	print $fh localtime(time) .' [message] ';
	print $fh @_;
	print $fh "\n";
	close $fh;
	ungensym($fh);
}

# === DEPRICATED ============================================================

#! DEPRICATED: lckey, ncget and ncset - use Puma::Util::NCKey instead
# --- get a case insensitive key from a hash ---
sub lckey { return ncget(@_); } # depricated
sub ncget {
   my ($hash, @keys) = @_;
	return undef unless ref $hash && @keys;
	foreach my $key (@keys) {
		$key = lc($key);
		foreach my $hashkey (keys %{$hash}) {
			return $hash->{$hashkey} if lc($hashkey) eq $key;
		}
	}
	return undef;
}
sub ncset {
   my ($hash, %args) = @_;
	foreach my $key (keys %args) {
		my $lckey = lc($key);
		foreach my $hashkey (keys %$hash) {
			delete $hash->{$hashkey} if lc($hashkey) eq $lckey;
		}
		$hash->{$key} = $args{$key};
	}
}

#! DEPRICATED: fixpath - use Puma::Util::FileSystem::cleanPath instead
sub fixpath {
	return undef unless @_;
	my $file = $_[0];
	if (@_ > 1) {
		if ($_[1] =~ /^\//) { $file = join '', @_;
		} else { $file = $_[1]; }
	}
	$file =~ s|/+|/|g;
	return $file;
}

# --- function used to remove keys from a hash that have a '' value ---
#! MOVED: now in ServerPages::Global
sub cleanhash {
	my %hash = @_;
	foreach my $key (keys %hash) {
		if ($key eq '') {
			delete $hash{$key};
		}
	}
	return %hash;
}

=head1 COPYRIGHT

Copyright (c) 2001, 2002 by Ingmar Ellenberger.

Distributed under The Artistic License.  For the text of this license,
see http://puma.site42.com/license.psp or read the file LICENSE in the
root of the distribution.

=cut

1;
