package Puma::Util::FileSystem;
use ibase Exporter;

@EXPORT_OK = qw( cleanPath ls filestat );

use Symbol;

use strict;

=function

NAME
	ls - return a hash containing a detailed file listing for a directory

SYNOPSIS

	use Puma::Util::FileSystem qw( ls );

	my $dirlist = ls('/usr/local/bin', '-r', '.*pl$');

DESCRIPTION

	The ls function returns to the caller a hash of information similar to the
	data returned by the Unix ls command.

PARAMETERS

	This takes the following positional parameters:

	$path - the path of the directory to be queried
	$params - various parameters and flags
		The only parameter that presently works is '-r' for recursion.
	$filter - a perl regex to filter the results by

=cut

sub ls {
	my ($path, $params, $filter) = (shift, shift || '', shift);
	my $recurse = $params =~ /r/ ? 1 : 0;
	$filter = '^[^\.]' unless defined $filter;

	my $list = {};

	my $dh = gensym(); # create dirhandle via Symbol::gensym
	opendir $dh, $path;
	foreach my $file (readdir $dh) {
		next unless $file =~ /$filter/;
		my $type = (-d "$path/$file") ? 'dir' : 'file';
		$list->{$file} = {
			name => $file,
			path => $path,
			type => $type,
			filestat($file)
		};

		next if $file eq '.' || $file eq '..';
		$list->{$file}->{content} = ls("$path/$file", $params, $filter)
			if $type eq 'dir' && $recurse;
	}
	ungensym($dh); # destroy dirhandle via Symbol::ungensym

	return $list;
}

sub filestat {
	my $file = shift;
	return unless $file;

	my @statarray = stat $file;

	my $stathash = {
		dev       => $statarray[0],
		inode     => $statarray[1],
		mode      => $statarray[2],
		numlink   => $statarray[3],
		uid       => $statarray[4],
		gid       => $statarray[5],
		rdev      => $statarray[6],
		size      => $statarray[7],
		atime     => $statarray[8],
		mtime     => $statarray[9],
		ctime     => $statarray[10],
		blocksize => $statarray[11],
		blocks    => $statarray[12],
	};

	return wantarray ? %$stathash : $stathash;
}

# === Functions for doing funky shit with paths =============================

# --- cleans up a filesystem path ---
sub cleanPath {
	my $path = shift;

	# --- remove all multiple '/'s ---
	$path =~ s|/+|/|g;

	# --- remove references the dir '.' ---
	undef while $path =~ s|/\./|/|g; # within the path
	$path =~ s|/\.$||g;              # at end of path

	# --- remove references the dir '..' ---
	undef while $path =~ s|/[^\./]+/\.\.||;

	return $path;
}

1;
