package Puma::Util::XML;
use ibase qw( Puma::Parser::XML Exporter );

=copyright

	Puma - Perl Universal Markup
	Copyright (c) 2001, Ingmar Ellenberger.  All rights reserved.
	Software is distributed AS IS.  No warranties, expressed or implied.

	This license is temporary, pending the selection of an appropriate
	open-source license.

=description

	XML.pm - A XML file parser based on Puma::Parser::XML

=cut

@EXPORT_OK = qw( xml2hash hash2xml );

use Puma::Util qw( readfile writefile );

use strict;

=TODO
	- eventually standardize the naming convention for packages
		- use package name and translate to appname
		- use appname as the standard?
=cut

sub parse {
   my $self = shift;

	my $ptree = $self->SUPER::parse(@_,
		XMLHandlers => {
			Init  => \&init,
			Start => \&start,
			Char  => \&char,
			End   => \&end,
			Final => \&final,
		},
		Object => $self,
	);

	return $ptree;
}

# === XML Handlers ==========================================================

sub init {
	my $self = shift;
	$self->{stack} = [];
	$self->{depth} = 0;
}

sub start {
	my ($self, $tagname) = (shift, shift);
	my $depth = $self->{depth}++;

	$self->{stack}->[$depth]->{TaGlAbEl} = $tagname;
	my %args = @_;
	foreach my $key (keys %args) {
		$self->addParam($key, $args{$key});
	}
}

sub char {
	my ($self, $content) = (shift, shift);
	return unless $self->{depth} > 0;

	# --- trim out leading and trailing whitespace ---
	$content =~ s/^\s*(.*?)\s*$/$1/g;
	return unless $content;
	$self->addParam($content, '');
}

sub end {
	my $self = shift;
	return unless $self->{depth} > 0;
	my $depth = --$self->{depth};

	my $key = $self->{stack}->[$depth]->{TaGlAbEl};
	my $value = $self->{stack}->[$depth + 1];

	$value = (keys %{$value})[0] if keys %{$value} == 1 && (values %{$value})[0] eq '';

	$self->addParam($key, $value);

	delete $self->{stack}->[$depth]->{TaGlAbEl};
	$self->{stack}->[$depth + 1] = {};
}

sub final {
	my $self = shift;
	$self->{hash} = $self->{stack}->[0]->{':parser'};
}

# === XML Handler Utilities ==================================================

sub addParam {
	my ($self, $key, $value) = @_;

	my $hash = $self->{stack}->[$self->{depth}];

	if (exists $hash->{$key} && defined $hash->{$key}) {
		if (ref $hash->{$key} eq 'ARRAY') {
			push @{$hash->{$key}}, $value;
		} else {
			$hash->{$key} = [ $hash->{$key}, $value ];
		}
	} else {
	   $hash->{$key} = $value;
	}

	$self->{stack}->[$self->{depth}] = $hash;
}


# === XML Generation (under development) ====================================

sub render {
	my ($self, $xmlhash) = @_;

	# --- get the xml hash (used for non-object calls) ---
	$self->{hash} = $xmlhash if $xmlhash;
	# --- start with an empty line buffer ---
	$self->{_buffer} = [ '' ];

	# --- call dump for each of the root level keys ---
	foreach my $key (sort keys %{$self->{hash}}) {
		$self->_dump($key, $self->{hash}->{$key}, 0);
	}	

	# --- build the xml from the line buffer ---
	my $xml = '';
	foreach my $line (@{$self->{_buffer}}) {
		$xml .= $line ."\n";
	}

	# --- add the header and return the generated xml ---
	return qq[<?xml version="1.0" encoding="UTF-8"?>\n$xml];
}

sub _dump {
	my ($self, $tag, $data, $depth) = @_;

	# --- deal with simple tags ---
	# <tag param/> or <tag>data</tag>
	return $self->add(
		flags => 'w', depth => $depth,
		text => $data =~ /^[\w.-]+$/ ? "<$tag $data/>" : "<$tag>$data</$tag>",
	) unless ref $data;

	# --- split hash into params, body and inner ---
	my (@params, $body, %inner);
	foreach my $key (sort keys %$data) {
		my $value = $data->{$key};

		# --- found an inner tag ---
		# <tag><innertag ... ></tag>
		if (ref $value) {
			$inner{$key} = $value;
			next;
		}

		if ($value eq '') {
			# --- found body or standalone param ---
			if ($key =~ /^[\w.-]+$/) {
				# --- found standalone param ---
				# <tag ... param ...>
				push @params, $key;
			} else {
				# --- found body ---
				# <tag ... >body</tag>
				$body = $key;
			}
		} else {
			# --- foud key/value pair ---
			if ($value =~ /\n/s) {
				# --- value has newlines, treat as inner tag ---
				# <tag><key>value</key></tag>
				$inner{$key} = $value;
			} else {
				if (lc($key) eq 'name') {
					unshift @params, qq[$key="$value"];
				} else {
					push @params, qq[$key="$value"];
				}
			}
		}
	}

	# --- create the start tag ---
	$self->add(
		flags => 'ns',
		depth => $depth,
		text  => [ "<$tag", @params ],
	);
	$self->add(
		flags => 'a',
		depth => $depth,
		text  => $body || %inner ? ">" : "/>",
		separator => '',
	);
	return unless $body || %inner;

	# --- add body ---
	$self->add(
		flags => 'w',
		depth => $depth + 1,
		text  => $body,
	) if $body;

	# --- inner stuff ---
	foreach my $key (sort keys %inner) {
		my $value = $inner{$key};

		if (ref($value) eq 'ARRAY') {
			foreach my $element (@$value) {
				$self->_dump($key => $element, $depth + 1);
			}
		} elsif (ref($value) eq 'HASH') {
			$self->_dump($key => $value, $depth + 1);
		} else {
			$self->add(
				flags => 'w',
				depth => $depth + 1,
				text  => "<$key>$value</$key>",
			);
		}
	}

	$self->add(
		flags => 'w',
		depth => $depth,
		text  => "</$tag>",
	);

	return;
}

=method
	flags => 'aensw'
		'a' - append to last line
		'e' - end line on completion
		'n' - new line
		's' - split between text
		'w' - whole line
	depth => indent depth
	separator => separator between text
	text => 'text' or [ 'text', 'text', ... ]
=cut

sub add {
	my ($self, %args) = @_;

	my $buffer = $self->{_buffer};
	my $maxlen = $self->{_maxlinelen} || 77;

	my $flags     = $args{flags}     || 'as';
	my $depth     = $args{depth}     || 0;
	my $separator = defined $args{separator} ? $args{separator} : ' ';
	my $text      = ref $args{text} ? $args{text} : [ $args{text} ];

	# --- remove excess spaces ---
	pop @$buffer if $buffer->[@$buffer - 1] eq '';

	my $line = '';
	# --- grab last if appending ---
	$line = pop @$buffer if $flags =~ /a/;

	foreach my $part (@$text) {
		# --- indent and and next if blank ---
		if ($line eq '') {
			$line = $self->_indent($depth) . $part;
			next;
		}

		my $lineplus = $line . $separator . $part;
		if ($flags =~ /s/) {
			if (length($lineplus) > $maxlen) {
				push @$buffer, $line;
				$line = $self->_indent($depth + 1) . $part;
			} else {
				$line = $lineplus;
			}
		} else {
			$line = $lineplus;
		}
	}

	push @$buffer, $line if $line;
	push @$buffer, '' if $flags =~ /[ew]/;

	return;
}

sub _indent {
	my ($self, $depth) = @_;
	my $indent = $self->{_indent} || '   ';
	return $indent x $depth;
}

# === exported methods ======================================================

use Puma::Util::NCHash;

sub xml2hash {
	my $args = nchash(@_);

	# --- at least one parameter required ---
	my $text = $args->{Text} || $args->{XML} || '';
	my $file = $args->{File} || $args->{FileName} || '';
	return unless $text || $file;

	# --- load and/or parse the XML ---
	my $xml = new Puma::Util::XML;
	if ($file) {
		$xml->parse(FileName => $file);
	} else {
		$xml->parse(Text => $text);
	}

	# --- return the hash ---
	return $xml->{hash};
}

sub hash2xml {
	my $args = nchash(@_);

	# --- get required parameters ---
	my $hash = $args->{Hash} || $args->{XML} || '';
	my $file = $args->{File} || $args->{FileName} || '';
	return unless $hash;

	# --- instantiate the parser ---
	my $xml = new Puma::Util::XML;

	# --- get optional parameters ---
	$xml->{_indent}     = $args->{Indent}    if defined $args->{Indent};
	$xml->{_maxlinelen} = $args->{MaxLength} if defined $args->{MaxLength};
	#! TODO: optional parameter - sortby

	my $text = $xml->render($hash);

	writefile($file, $text) if $file;

	return $text
}

1;
