package Puma::Devel::Form;
use ibase Puma::Tag;

use Puma::Devel::FormClient;
use Puma::Util qw[ ncget ncset splithash ];
use Puma::Util::NCHash;

use strict;

# --- construct the form object, doing any submit processisg as necessary ---
sub construct {
	my $self = shift;

	# --- grab :use tag parameters ---
	my %args = $self->SUPER::construct(@_);
	$self->stripArgs(\%args, qw[
		Application Bind
		Server
		Name
	]);

	# --- initialize object vars ---
	my $app    = $self->application($self->{Application} || $self->{Bind});
	my $server = $self->server($self->{Server});
	my $name   = $self->name($self->{Name});

	# --- set the submitted flag ---
	$self->submitted(1) if ($server->param('_FormName') || '') eq $name;

	# --- load up the application object (submit processing) ---
	$self->loadApplication;

	return %args;
}

# === Methods for the Application object ====================================

# --- load the app object with form data ---
sub loadApplication {
	my $self = shift;
	my $app = $self->application;

	# --- push CGI params to app object if submit ---
	if ($self->submitted) {
		my $params = $self->server->paramHash;
		foreach my $key (keys %$params) {
			$self->setFormValue($key, $params->{$key});
		}
	}

	# --- allow the application to do form validation/initialization ---
	my $valid = 1;
	$valid = $app->initForm(Form => $self, @_)
		if ref $app ne 'HASH' && $app->can('initForm');

	return $self->valid($valid);
}

# --- get form values from application ---
sub getFormValue {
	my ($self, $key) = @_;
	my $app = $self->application;

	# --- use app->getFormValue if it exists ---
	return $app->getFormValue($key)
		if ref $app ne 'HASH' && $app->can('getFormValue');
	return $app->{$key} || '';
}

# --- set application values from form ---
sub setFormValue {
	my ($self, $key, $value) = @_;
	my $app = $self->application;

	# --- use app->setFormValue if it exists ---
	return $app->setFormValue($key, $value)
		if ref $app ne 'HASH' && $app->can('setFormValue');
	return $app->{$key} = $value;
}

# === Accessors =============================================================

# --- the bound application ---
sub application {
	my ($self, $app) = @_;
	$self->{_application} = $app if defined $app;
	return $self->{_application} ||= new Puma::Devel::FormClient();
}

# --- the server object ---
sub server {
	my ($self, $app) = @_;
	$self->{_server} = $app if defined $app;
	return $self->{_server} ||= new Puma::Devel::FormClient();
}

# --- form name ---
sub name {
	my ($self, $name) = @_;
	$self->{_name} = $name if defined $name;
	return $self->{Name} || '__unknown__';
}

# --- submitted flag ---
sub submitted {
	my ($self, $submit) = @_;

	$self->{_submit} = $submit if defined $submit;
	return $self->{_submit} || 0;
}

# --- validated flag ---
sub valid {
	my ($self, $valid) = @_;

	$self->{_valid} = $valid if defined $valid;
	return $self->{_valid} || 0;
}

# === Form tags =============================================================

sub form { my $self = shift; $self->_form(@_); }

sub button { shift->_button('button', @_); }
sub reset  { shift->_button('reset',  @_); }
sub submit { shift->_button('submit', @_); }
sub image  { shift->_button('image',  @_); }

sub file     { shift->_input('file',     @_); }
sub hidden   { shift->_input('hidden',   @_); }
sub password { shift->_input('password', @_); }
sub text     { shift->_input('text',     @_); }

sub checkbox { shift->_checkbox('checkbox', @_); }
sub radio    { shift->_checkbox('radio',    @_); }

sub select   { my $self = shift; $self->_select(@_); }
sub option   { my $self = shift; $self->_option(@_); }
sub textarea { my $self = shift; $self->_textarea(@_); }

# === Private methods to generate form tags =================================

sub _form {
	my ($self, $args) = (shift, nchash(@_));

	# --- grab the parameters ---
	$self->{onSuccess} = delete $args->{onSuccess};
	$args->{name} = $self->name($args->{name});

	# --- opening tag ---
	print '<form'. $self->paramString(%$args) .'>';
	print qq[<input type="hidden" name="_FormName" value="$args->{name}">];

	# --- call the app methods and draw the children ---
#	$self->_startForm(
#		Form => $self,
#		FormName => $name,
#	);
	$self->renderChildren();
	#! TODO: add basic validation (i.e. required fields)
#	$self->_endForm(
#		Form => $self,
#		FormName => $name,
#		Validate => 1,
#		Submit   => $self->submitted,
#	);

	# --- closing tag ---
	print '</form>';

	# --- redirect if sucessful submit ---
	$self->server->redirect($self->{onSuccess})
		if $self->submitted && $self->valid == 1;

	# --- clean thing up and return ---
	return;
}

sub _button {
	my ($self, $type, $args) = (shift, shift, nchash(@_));
	$args->{type} = $type;
	print '<input'. $self->paramString(%$args) .'>';
}

sub _input {
	my ($self, $type, $args) = (shift, shift, nchash(@_));
	$args->{type} = $type;

	# --- get value param from application ---
	$args->{value} = $self->getFormValue($args->{name})
		if defined delete $args->{appvalue};

	print '<input'. $self->paramString(%$args) .'>';
}

sub _checkbox {
	my ($self, $type, $args) = (shift, shift, nchash(@_));
	$args->{type} = $type;

	# --- get value param from application ---
	if (defined delete $args->{appvalue}) {
		$args->{checked} = ''
			if $args->{value} eq ($self->getFormValue($args->{name}) || '');
		#print "<br>'$args->{value}' '$args->{name}' '". ($self->getFormValue($args->{name}) || '?') ."'<br>\n";
	}

	print '<input'. $self->paramString(%$args) .'>';
}

sub _select {
	my ($self, %args) = @_;
	print "<select". $self->paramString(%args) .">";
	$self->renderChildren();
	print "</select>";
}

sub _option {
	my ($self, %args) = @_;
	print "<option". $self->paramString(%args) .">";
	$self->renderChildren();
	print "</option>";
}

sub _textarea {
	my ($self, $args) = (shift, nchash(@_));

	# --- get value param from application ---
	my $value = delete $args->{value} || '';
	$value = $self->getFormValue($args->{name}) || ''
		if defined delete $args->{appvalue};

	# --- build the tag and render the children ---
	print '<textarea'. $self->paramString(%$args) .">$value";
	$self->renderChildren();
	print "</textarea>";
}

# === Utility Methods =======================================================

# --- convert a hash to a string of parameters ---
sub paramString {
	my ($self, %params) = @_;

	my $paramstr;
	foreach my $key (keys %params) {
		my $value = $params{$key};
		$value = '' unless defined $value;
		if ($value eq '' && $key =~ /checked|selected/) {
			$paramstr .= " $key";
		} else {
			$paramstr .= " $key=\"$value\"";
		}
	}
	return $paramstr;
}



1;

__END__



sub _stdProc {
	my ($self, %args) = @_;

	my $tag = splithash(\%args, qw[ appvalue required ]);

	# --- get value param from application ---
	ncset(\%args, value => $self->getFormValue(ncget(\%args, 'name')))
		if exists $tag->{appvalue} && defined $tag->{appvalue};

#	# --- submit processing ---
#	if ($self->submitted) {
#		my $name  = ncget(\%args, 'name');
#		my $value = ncget(\%args, 'value');
#		if (defined $tag->{required} && 
#		    (!defined $value || $value eq '')) {
#			$self->valid(0);
#			$self->error("'$name' field required");
#		}
#	}

	return %args;
}


# === FormClient Interfaces =================================================

sub _startForm {
	my $self = shift;
	my $app = $self->{'_application'};

	return $app->startForm(@_)
		if ref $app ne 'HASH' && $app->can('startForm');
}

sub _endForm {
	my $self = shift;
	my $app = $self->{'_application'};

	return $app->endForm(@_)
		if ref $app ne 'HASH' && $app->can('endForm');
}

# === Error handling routines ===============================================

sub error {
	my ($self, $message) = @_;
	$self->{_error} ||= [];

	push @{$self->{_error}}, $message if defined $message;
	return $self->{_error};
}

sub errorMessages {
	my $self = shift;
	# TODO: return a constructed array when the full error structure is defined
	return $self->{_error};
}

1;
